# Copyright (c) 2025 Tom Callister
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
import warnings
from typing import Callable, Dict, Optional, Sequence, Tuple

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jrd
from astropy import units
from astropy.cosmology import Planck15, z_at_value
from jaxtyping import Array, PRNGKeyArray

from ...parameters import Parameters
from ...utils.tools import error_if
from ._emulator import Emulator


class pdet_O3(Emulator):
    """Class implementing the LIGO-Hanford, LIGO-Livingston, and Virgo network's
    selection function during their O3 observing run.

    Used to evaluate the detection probability of compact binaries, assuming a false
    alarm threshold of below 1 per year. The computed detection probabilities include
    all variation in the detectors' sensitivities over the course of the O3 run and
    accounts for time in which the instruments were not in observing mode. They should
    therefore be interpreted as the probability of a CBC detection if that CBC occurred
    during a random time between the startdate and enddate of O3.
    """

    all_parameters: Sequence[str] = eqx.field(static=True)
    proposal_dist: Dict[str, Tuple[float, float]] = eqx.field(static=True)
    parameters: Sequence[str] = eqx.field(static=True)
    extra_shape: Tuple[int, ...] = eqx.field(static=True)
    interp_DL: Array = eqx.field(static=False)
    interp_z: Array = eqx.field(static=False)
    is_full: bool = eqx.field(static=True)

    def __init__(
        self,
        model_weights=None,
        scaler=None,
        parameters: Optional[Sequence[str]] = None,
        batch_size: Optional[int] = None,
        scale: float = 1.0,
    ):
        """Instantiates a `p_det_O3` object, subclassed from the `emulator` class.

        Parameters
        ----------
        model_weights : `None` or `str`
            Filepath to .hdf5 file containing trained network weights, as saved
            by a `tensorflow.keras.Model.save_weights`, command, if one wishes
            to override the provided default weights (which are loaded when
            `model_weights==None`).
        scaler : `str`
            Filepath to saved `sklearn.preprocessing.StandardScaler` object, if
            one wishes to override the provided default (loaded when
            `scaler==None`).
        parameters : `list`
            List of parameters to be used by the emulator.
        batch_size : `int`
            Batch size to be used by `jax.vmap`
        """

        error_if(parameters is None, msg="Must provide list of parameters")

        self.all_parameters = (
            Parameters.PRIMARY_MASS_SOURCE.value,
            Parameters.SECONDARY_MASS_SOURCE.value,
            Parameters.PRIMARY_SPIN_MAGNITUDE.value,
            Parameters.SECONDARY_SPIN_MAGNITUDE.value,
            Parameters.COS_TILT_1.value,
            Parameters.COS_TILT_2.value,
            Parameters.PHI_12.value,
            Parameters.REDSHIFT.value,
            Parameters.COS_IOTA.value,
            Parameters.POLARIZATION_ANGLE.value,
            Parameters.RIGHT_ASCENSION.value,
            Parameters.SIN_DECLINATION.value,
        )

        error_if(
            Parameters.PRIMARY_MASS_SOURCE.value not in parameters,
            msg="Must include {0} parameter".format(
                Parameters.PRIMARY_MASS_SOURCE.value
            ),
        )
        error_if(
            Parameters.SECONDARY_MASS_SOURCE.value not in parameters,
            msg="Must include {0} parameter".format(
                Parameters.SECONDARY_MASS_SOURCE.value
            ),
        )

        self.is_full = set(self.all_parameters) == set(parameters)
        self.proposal_dist = {}
        if not self.is_full:
            # if not every parameter is provided, we are only storing the bounds of the
            # missing parameters
            missing_params = set(self.all_parameters) - set(parameters)

            proposal_dist = {
                Parameters.REDSHIFT.value: (0.0, 10.0),
                Parameters.COS_IOTA.value: (-1.0, 1.0),
                Parameters.POLARIZATION_ANGLE.value: (0.0, jnp.pi),
                Parameters.RIGHT_ASCENSION.value: (0.0, 2.0 * jnp.pi),
                Parameters.SIN_DECLINATION.value: (-1.0, 1.0),
            }
            for param in missing_params:
                if param in proposal_dist:
                    self.proposal_dist[param] = proposal_dist[param]

        self.parameters = parameters

        self.extra_shape = (100,)

        if model_weights is None:
            model_weights = os.path.join(
                os.path.dirname(__file__), "./weights_HLV_O3.hdf5"
            )
        else:
            warnings.warn("Overriding default weights", UserWarning)

        if scaler is None:
            scaler = os.path.join(os.path.dirname(__file__), "./scaler_HLV_O3.json")
        else:
            warnings.warn("Overriding default weights", UserWarning)

        input_dimension = 15
        hidden_width = 192
        hidden_depth = 4
        activation = lambda x: jax.nn.leaky_relu(x, 1e-3)
        final_activation = lambda x: (1.0 - 0.0589) * jax.nn.sigmoid(x)

        self.interp_DL = jnp.logspace(-4, jnp.log10(15.0), 500)
        self.interp_z = z_at_value(
            Planck15.luminosity_distance,
            self.interp_DL * units.Gpc,
        ).value

        super().__init__(
            model_weights,
            scaler,
            input_dimension,
            hidden_width,
            hidden_depth,
            activation,
            final_activation,
            batch_size,
            scale,
        )

    def _transform_parameters(
        self,
        m1_trials: Array,
        m2_trials: Array,
        a1_trials: Array,
        a2_trials: Array,
        cost1_trials: Array,
        cost2_trials: Array,
        phi12_trials: Array,
        z_trials: Array,
        cos_iota_trials: Array,
        pol_trials: Array,
        ra_trials: Array,
        sin_dec_trials: Array,
    ) -> Array:
        q = m2_trials / m1_trials
        eta = q / (1.0 + q) ** 2
        Mtot_det = (m1_trials + m2_trials) * (1.0 + z_trials)
        Mc_det = eta ** (3.0 / 5.0) * Mtot_det

        DL = jnp.interp(z_trials, self.interp_z, self.interp_DL)
        Mc_DL_ratio = Mc_det ** (5.0 / 6.0) / DL
        amp_factor_plus = jnp.log((Mc_DL_ratio * ((1.0 + cos_iota_trials**2) / 2)) ** 2)
        amp_factor_cross = jnp.log((Mc_DL_ratio * cos_iota_trials) ** 2)

        # Effective spins
        chi_effective = (a1_trials * cost1_trials + q * a2_trials * cost2_trials) / (
            1.0 + q
        )
        chi_diff = (a1_trials * cost1_trials - a2_trials * cost2_trials) / 2.0

        # Generalized precessing spin
        Omg = q * (3.0 + 4.0 * q) / (4.0 + 3.0 * q)
        chi_1p = a1_trials * jnp.sqrt(1.0 - cost1_trials**2)
        chi_2p = a2_trials * jnp.sqrt(1.0 - cost2_trials**2)
        chi_p_gen = jnp.sqrt(
            chi_1p**2
            + (Omg * chi_2p) ** 2
            + 2.0 * Omg * chi_1p * chi_2p * jnp.cos(phi12_trials)
        )

        return jnp.stack(
            [
                amp_factor_plus,
                amp_factor_cross,
                Mc_det,
                Mtot_det,
                eta,
                q,
                DL,
                ra_trials,
                sin_dec_trials,
                jnp.abs(cos_iota_trials),
                jnp.sin(pol_trials % jnp.pi),
                jnp.cos(pol_trials % jnp.pi),
                chi_effective,
                chi_diff,
                chi_p_gen,
            ],
            axis=-1,
        )

    def check_input(
        self,
        key: PRNGKeyArray,
        shape: tuple[int, ...],
        parameter_dict: dict[str, Array],
    ) -> Tuple[PRNGKeyArray, dict[str, Array]]:
        """Method to check provided set of compact binary parameters for any missing
        information, and/or to augment provided parameters with any additional derived
        information expected by the neural network. If extrinsic parameters (e.g. sky
        location, polarization angle, etc.) have not been provided, they will be
        randomly generated and appended to the given CBC parameters.

        Parameters
        ----------
        parameter_dict : `dict` or `pd.DataFrame`
            Set of compact binary parameters for which we want to evaluate pdet

        Returns
        -------
        parameter_dict : `dict`
            Dictionary of CBC parameters, augmented with necessary derived
            parameters
        """

        missing_params = {}

        if Parameters.PRIMARY_SPIN_MAGNITUDE.value not in parameter_dict:
            missing_params[Parameters.PRIMARY_SPIN_MAGNITUDE.value] = jnp.zeros(shape)

        if Parameters.SECONDARY_SPIN_MAGNITUDE.value not in parameter_dict:
            missing_params[Parameters.SECONDARY_SPIN_MAGNITUDE.value] = jnp.zeros(shape)

        if Parameters.COS_TILT_1.value not in parameter_dict:
            missing_params[Parameters.COS_TILT_1.value] = jnp.ones(shape)

        if Parameters.COS_TILT_2.value not in parameter_dict:
            missing_params[Parameters.COS_TILT_2.value] = jnp.ones(shape)

        if Parameters.PHI_12.value not in parameter_dict:
            missing_params[Parameters.PHI_12.value] = jnp.zeros(shape)

        parameter_dict.update(missing_params)

        return key, parameter_dict

    def predict(self, key: PRNGKeyArray, params: Array) -> Array:
        shape = jnp.shape(params)[:-1]
        if self.is_full:
            parameter_dict = {
                parameter: jax.lax.dynamic_index_in_dim(
                    params, i, axis=-1, keepdims=False
                )
                for i, parameter in enumerate(self.parameters)
            }

            @jax.jit
            def _predict(_features: Array) -> Array:
                prediction = jnp.squeeze(self.__call__(_features), axis=-1)
                prediction = jnp.nan_to_num(
                    prediction, nan=0.0, posinf=jnp.inf, neginf=-jnp.inf
                )
                return prediction

            features = jnp.stack(
                [parameter_dict[param] for param in self.all_parameters],
                axis=-1,
            )
        else:
            parameter_dict = {
                parameter: jnp.broadcast_to(
                    jax.lax.dynamic_index_in_dim(params, i, axis=-1, keepdims=False),
                    self.extra_shape + shape,
                )
                for i, parameter in enumerate(self.parameters)
            }
            key, parameter_dict = self.check_input(
                key, self.extra_shape + shape, parameter_dict
            )
            for param in self.proposal_dist.keys():
                parameter_dict[param] = jrd.uniform(
                    key,
                    self.extra_shape + shape,
                    minval=self.proposal_dist[param][0],
                    maxval=self.proposal_dist[param][1],
                )

            @jax.jit
            def _predict(_features: Array) -> Array:
                prediction = jnp.squeeze(self.__call__(_features), axis=-1)
                prediction = jnp.nan_to_num(
                    prediction, nan=0.0, posinf=jnp.inf, neginf=-jnp.inf
                )
                return jnp.mean(prediction, axis=0)

            features = jnp.stack(
                [parameter_dict[param] for param in self.all_parameters],
                axis=-1,
            )
            features = jnp.moveaxis(features, 1, 0)

        return jax.lax.map(_predict, features, batch_size=self.batch_size)

    def get_logVT(self) -> Callable[[Array], Array]:
        raise NotImplementedError("logVT is not implemented for pdet_O3")

    def get_mapped_logVT(self) -> Callable[[Array], Array]:
        return lambda x: jnp.log(self.scale * self.predict(jrd.PRNGKey(111), x))



# Copyright Contributors to the Pyro project.
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Sequence
from typing import Tuple

import jax
import jax.numpy as jnp
import jax.random as jrd
import numpyro.distributions as dist
import pytest
from jaxtyping import Array
from numpyro.distributions import Distribution, HalfNormal, MultivariateNormal, Normal

from gwkokab.models.utils import ScaledMixture


rng_key = jax.random.PRNGKey(42)


def get_normal(batch_shape: Tuple[int, ...]) -> Normal:
    """Get parameterized Normal with given batch shape."""
    loc = jnp.zeros(batch_shape)
    scale = jnp.ones(batch_shape)
    normal = Normal(loc=loc, scale=scale)
    return normal


def get_half_normal(batch_shape: Tuple[int, ...]) -> HalfNormal:
    """Get parameterized HalfNormal with given batch shape."""
    scale = jnp.ones(batch_shape)
    half_normal = HalfNormal(scale=scale)
    return half_normal


def get_mvn(batch_shape: Tuple[int, ...]) -> MultivariateNormal:
    """Get parameterized MultivariateNormal with given batch shape."""
    dimensions = 2
    loc = jnp.zeros((*batch_shape, dimensions))
    cov_matrix = jnp.eye(dimensions, dimensions)
    for i, s in enumerate(batch_shape):
        loc = jnp.repeat(jnp.expand_dims(loc, i), s, axis=i)
        cov_matrix = jnp.repeat(jnp.expand_dims(cov_matrix, i), s, axis=i)
    mvn = MultivariateNormal(loc=loc, covariance_matrix=cov_matrix)
    return mvn


@pytest.mark.parametrize("jax_dist_getter", [get_normal, get_mvn])
@pytest.mark.parametrize("nb_mixtures", [1, 3])
@pytest.mark.parametrize("batch_shape", [(), (1,), (7,), (2, 5)])
def test_mixture_same_batch_shape(jax_dist_getter, nb_mixtures, batch_shape):
    log_scales = jrd.normal(rng_key, (nb_mixtures,)) * jnp.pi
    for i, s in enumerate(batch_shape):
        log_scales = jnp.repeat(jnp.expand_dims(log_scales, i), s, axis=i)
    component_distribution = [jax_dist_getter(batch_shape) for _ in range(nb_mixtures)]
    _test_mixture(log_scales, component_distribution)


@pytest.mark.parametrize("jax_dist_getter", [get_normal, get_mvn])
@pytest.mark.parametrize("nb_mixtures", [3])
@pytest.mark.parametrize("mixing_batch_shape, component_batch_shape", [[(2,), (7, 2)]])
def test_mixture_broadcast_batch_shape(
    jax_dist_getter, nb_mixtures, mixing_batch_shape, component_batch_shape
):
    # Create mixture
    log_scales = jrd.normal(rng_key, (nb_mixtures,)) * jnp.pi
    for i, s in enumerate(mixing_batch_shape):
        log_scales = jnp.repeat(jnp.expand_dims(log_scales, i), s, axis=i)
    component_distribution = [
        jax_dist_getter(component_batch_shape) for _ in range(nb_mixtures)
    ]
    _test_mixture(log_scales, component_distribution)


@pytest.mark.parametrize("batch_shape", [(), (1,), (7,), (2, 5)])
@pytest.mark.filterwarnings(
    "ignore:Out-of-support values provided to log prob method."
    " The value argument should be within the support.:UserWarning"
)
def test_mixture_with_different_support(batch_shape):
    log_scales = jrd.normal(rng_key, (2,)) * jnp.pi
    component_distribution = [
        get_normal(batch_shape),
        get_half_normal(batch_shape),
    ]
    mixture = ScaledMixture(
        log_scales=log_scales,
        component_distributions=component_distribution,
        support=dist.constraints.real,
    )
    assert mixture.batch_shape == batch_shape
    sample_shape = (11,)
    component_distribution[0]._validate_args = True
    component_distribution[1]._validate_args = True
    xx = component_distribution[0].sample(rng_key, sample_shape)
    log_prob_0 = component_distribution[0].log_prob(xx)
    log_prob_1 = component_distribution[1].log_prob(xx)
    expected_log_prob = jax.scipy.special.logsumexp(
        jnp.stack([log_prob_0 + log_scales[0], log_prob_1 + log_scales[1]], axis=-1),
        axis=-1,
    )
    result = mixture.log_prob(xx)
    assert jnp.allclose(result, expected_log_prob)


def _test_mixture(log_scales: Array, component_distribution: Sequence[Distribution]):
    # Create mixture
    mixture = ScaledMixture(
        log_scales=log_scales,
        component_distributions=component_distribution,
    )
    assert mixture.mixture_size == log_scales.shape[-1], (
        "Mixture size needs to be the size of the probability vector"
    )

    if isinstance(component_distribution, dist.Distribution):
        assert mixture.batch_shape == component_distribution.batch_shape[:-1], (
            "Mixture batch shape needs to be the component batch shape without the mixture dimension."
        )
    else:
        assert mixture.batch_shape == component_distribution[0].batch_shape, (
            "Mixture batch shape needs to be the component batch shape."
        )
    # Test samples
    sample_shape = (11,)
    # Samples from component distribution(s)
    component_samples = mixture.component_sample(rng_key, sample_shape)
    assert component_samples.shape == (
        *sample_shape,
        *mixture.batch_shape,
        mixture.mixture_size,
        *mixture.event_shape,
    )
    # Samples from mixture
    samples = mixture.sample(rng_key, sample_shape=sample_shape)
    assert samples.shape == (*sample_shape, *mixture.batch_shape, *mixture.event_shape)
    # Check log_prob
    lp = mixture.log_prob(samples)
    nb_value_dims = len(samples.shape) - mixture.event_dim
    expected_shape = samples.shape[:nb_value_dims]
    assert lp.shape == expected_shape
    # Samples with indices
    samples_, [indices] = mixture.sample_with_intermediates(
        rng_key, sample_shape=sample_shape
    )
    assert samples_.shape == samples.shape
    assert indices.shape == (*sample_shape, *mixture.batch_shape)
    assert jnp.issubdtype(indices.dtype, jnp.integer)
    assert (indices >= 0).all() and (indices < mixture.mixture_size).all()
    # Check mean
    mean = mixture.mean
    assert mean.shape == mixture.shape()
    # Check variance
    var = mixture.variance
    assert var.shape == mixture.shape()
    # Check cdf
    if mixture.event_shape == ():
        cdf = mixture.cdf(samples)
        assert cdf.shape == (*sample_shape, *mixture.shape())



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import os

import mplcursors
import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    """Create the command line argument parser.

    This function creates the command line argument parser and returns it.

    Returns
    -------
    argparse.ArgumentParser
        the command line argument parser
    """
    parser = argparse.ArgumentParser(
        description="Scatter 3D plotter.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a 3D scatter plot.",
    )
    parser.add_argument(
        "-d",
        "--data",
        help="data file path. Only .dat files are supported.",
        required=True,
        type=argparse.FileType("r"),
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "-x",
        "--x-value-column-name",
        help="name of the x-axis values",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-y",
        "--y-value-column-name",
        help="name of the y-axis values",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-z",
        "--z-value-column-name",
        help="name of the z-axis values",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--title",
        help="title of the plot",
        type=str,
    )
    parser.add_argument(
        "-xl",
        "--xlabel",
        help="label of the x-axis",
        default="x",
        type=str,
    )
    parser.add_argument(
        "-yl",
        "--ylabel",
        help="label of the y-axis",
        default="y",
        type=str,
    )
    parser.add_argument(
        "-zl",
        "--zlabel",
        help="label of the z-axis",
        default="z",
        type=str,
    )
    parser.add_argument(
        "-cmap",
        "--color-map",
        help="color map for the plot",
        default="plasma",
        type=str,
    )
    parser.add_argument(
        "-c",
        "--color",
        help="index of the color values",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-cbar",
        "--color-bar",
        help="show color bar",
        action="store_true",
    )
    parser.add_argument(
        "-a",
        "--alpha",
        help="transparency of the points",
        default=0.3,
        type=float,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    data = pd.read_csv(args.data.name, delimiter=" ")
    x = data[args.x_value_column_name].to_numpy()
    y = data[args.y_value_column_name].to_numpy()
    z = data[args.z_value_column_name].to_numpy()

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")

    sc = ax.scatter(
        x,
        y,
        z,
        c=data[:, args.color] if args.color is not None else z,
        cmap=args.color_map,
        marker="o",
        alpha=args.alpha,
    )
    if args.title is not None:
        ax.set_title(args.title)
    ax.set_xlabel(args.xlabel)
    ax.set_ylabel(args.ylabel)
    ax.set_zlabel(args.zlabel)
    if args.color_bar:
        cbar = fig.colorbar(sc)
        cbar.set_label("Color")
        cursor = mplcursors.cursor(sc, hover=True)
        cursor.connect(
            "add",
            lambda sel: sel.annotation.set_text(
                f"{sel.annotation.get_text()}\n{sel.artist.get_array()}"
            ),
        )
    plt.tight_layout()
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")
    args.data.close()



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse

import corner
import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Corner plotter.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a corner plot.",
    )
    parser.add_argument(
        "-d",
        "--data",
        help="data file path. Only .dat files are supported.",
        nargs="+",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=str,
    )
    parser.add_argument(
        "--data-labels",
        help="labels for the data files",
        nargs="+",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-l",
        "--column-labels",
        nargs="+",
        help="labels for the corner plot",
        default=None,
    )
    parser.add_argument(
        "-t",
        "--truths",
        help="truth values for the corner plot",
        nargs="+",
        type=float,
    )
    parser.add_argument(
        "-q",
        "--quantiles",
        help="quantiles for the corner plot",
        default=None,
        type=list[float],  # type: ignore[arg-type]
    )
    parser.add_argument(
        "-b",
        "--bins",
        help="number of bins for the corner plot",
        default=20,
        type=int,
    )
    parser.add_argument(
        "-s",
        "--smooth",
        help="smooth the corner plot",
        action="store_true",
    )
    parser.add_argument(
        "-st",
        "--show-titles",
        help="show titles in the corner plot",
        action="store_true",
    )
    parser.add_argument(
        "-scale",
        "--scale",
        help="scale the corner plot",
        default=1.0,
        type=float,
    )
    parser.add_argument(
        "-size",
        "--size",
        help="size of the corner plot in inches",
        nargs=2,
        default=(6, 6),
        type=float,
    )
    parser.add_argument(
        "--range",
        help="range of the corner plot",
        nargs="+",
        action="append",
        default=None,
        type=float,
    )
    parser.add_argument(
        "--truth-color",
        help="color of the truth values in the corner plot",
        default="red",
        type=str,
    )
    parser.add_argument(
        "--colors",
        help="colors of the corner plot",
        nargs="+",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )
    parser.add_argument(
        "--title-fmt",
        help="format string for titles in the corner plot",
        default=".3f",
        type=str,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    N = len(args.data)
    if args.colors is None:
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        assert len(colors) >= N, (
            f"We could only get {len(colors)} default colors from matplotlib, but {N} "
            "data files were provided. Consider providing a list of colors."
        )
        colors = colors[:N]
    else:
        assert len(args.colors) == N, (
            "Number of colors must match number of data files."
        )
        colors = args.colors

    with open(args.data[0]) as f:
        header = f.readline().lstrip("#").strip().split()
    assert len(header) > 0, "Header must not be empty."

    if args.column_labels is None:
        column_labels = header
    else:
        column_labels = args.column_labels

    assert len(header) == len(column_labels), (
        f"Number of columns in the header ({len(header)}) does not match the number "
        f"of column labels provided ({len(column_labels)})."
    )

    data = []
    for file in args.data:
        data_i = pd.read_csv(file, delimiter=" ", skiprows=1).to_numpy()
        with open(file) as f:
            header_i = f.readline().lstrip("#").strip().split()
        assert len(header_i) == len(header), (
            f"Number of columns in {file} are not equal to the expected number {len(header)}."
        )
        reordered_indices = [header.index(col) for col in header_i]
        data_i = data_i[:, reordered_indices]

        data.append(data_i)

    scaling_factor = args.scale
    fig = plt.figure(
        figsize=(scaling_factor * args.size[0], scaling_factor * args.size[1]),
        dpi=args.dpi,
    )

    for i in range(N):
        fig = corner.corner(
            data[i],
            fig=fig,
            labels=column_labels,
            truths=args.truths if args.truths is not None else None,
            quantiles=args.quantiles,
            bins=args.bins,
            smooth=args.smooth,
            show_titles=args.show_titles,
            truth_color=args.truth_color,
            color=colors[i],
            plot_datapoints=False,
            range=args.range,
            title_fmt=args.title_fmt,
            plot_density=False,
            hist_kwargs={"density": True},
        )

    legend_handles = [plt.Line2D([], [], color=color, linewidth=2) for color in colors]
    fig.legend(
        legend_handles,
        args.data_labels,
        loc="upper right",
        # fontsize=15,
    )

    plt.savefig(args.output, bbox_inches="tight")
    plt.close("all")



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""
Provides implementation of various models, constraints, and transformations using
`NumPyro <https://github.com/pyro-ppl/numpyro>`_.
"""

from . import (
    constraints as constraints,
    transformations as transformations,
    utils as utils,
)
from ._models import (
    PowerlawPeak as PowerlawPeak,
    PowerlawPrimaryMassRatio as PowerlawPrimaryMassRatio,
    SmoothedTwoComponentPrimaryMassRatio as SmoothedTwoComponentPrimaryMassRatio,
    Wysocki2019MassModel as Wysocki2019MassModel,
)
from .npowerlawmgaussian import NPowerlawMGaussian as NPowerlawMGaussian
from .redshift import PowerlawRedshift as PowerlawRedshift
from .spin import (
    BetaFromMeanVar as BetaFromMeanVar,
    GaussianSpinModel as GaussianSpinModel,
    IndependentSpinOrientationGaussianIsotropic as IndependentSpinOrientationGaussianIsotropic,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
r"""During analysis, we often need to create various types of plots to visualize data or
results. The GWKokab core team has identified several common plot types based on
recurring data patterns. These plots may require input from a single file or multiple
files, depending on the specific requirements.

To streamline the process of generating these plots, we have developed a set of
command-line interfaces (CLIs) built on top of the following libraries:

- `matplotlib <https://matplotlib.org/>`_
- `seaborn <https://seaborn.pydata.org/>`_
- `corner <https://corner.readthedocs.io/>`_
- `arviz <https://python.arviz.org/en/stable/>`_

The available CLIs are as follows:

- :code:`gwk_batch_scatter2d`
- :code:`gwk_batch_scatter3d`
- :code:`gwk_chain_plot`
- :code:`gwk_corner_plot`
- :code:`gwk_ess_plot`
- :code:`gwk_joint_plot`
- :code:`gwk_ppd_plot`
- :code:`gwk_r_hat_plot`
- :code:`gwk_samples_from_vt`
- :code:`gwk_scatter2d`
- :code:`gwk_scatter3d`

For detailed usage instructions, use the help option :code:`-h` or :code:`--help` with
the respective CLI name.

We welcome feedback and community contributions to improve these CLIs. If you have
suggestions or wish to contribute, please refer to the :doc:`/contributing` guide for
more details.

.. note::

    The plots generated by these CLIs are primarily intended for testing purposes, so
    we have not prioritized aesthetic enhancements.
"""



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import argparse
import glob
import os

import pandas as pd
from matplotlib import pyplot as plt


def make_parser() -> argparse.ArgumentParser:
    """Create the command line argument parser.

    This function creates the command line argument parser and returns it.

    Returns
    -------
    argparse.ArgumentParser
        the command line argument parser
    """
    parser = argparse.ArgumentParser(
        description="Scatter 2D plotter.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="This script plots a 2D scatter plot.",
    )
    parser.add_argument(
        "-d",
        "--data-regex",
        help="regex pattern for the data files. Only .dat files are supported.",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-o",
        "--output",
        help="output file path",
        required=True,
        type=argparse.FileType("w"),
    )
    parser.add_argument(
        "-x",
        "--x-value-column-name",
        help="name of the x-axis values",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-y",
        "--y-value-column-name",
        help="name of the y-axis values",
        required=True,
        type=str,
    )
    parser.add_argument(
        "-t",
        "--title",
        help="title of the plot",
        default=None,
        type=str,
    )
    parser.add_argument(
        "-xl",
        "--xlabel",
        help="label of the x-axis",
        default="x",
        type=str,
    )
    parser.add_argument(
        "-yl",
        "--ylabel",
        help="label of the y-axis",
        default="y",
        type=str,
    )
    parser.add_argument(
        "-s",
        "--scatter-size",
        help="size of the scatter points",
        default=5,
        type=int,
    )
    parser.add_argument(
        "-a",
        "--alpha",
        help="transparency of the scatter points",
        default=0.3,
        type=float,
    )
    parser.add_argument(
        "-m",
        "--marker",
        help="marker style for the scatter points",
        default=".",
        type=str,
    )
    parser.add_argument(
        "--use-latex",
        help="use LaTeX for rendering text",
        action="store_true",
    )
    parser.add_argument(
        "--font-family",
        help="font family to use",
        type=str,
        default=None,
    )
    parser.add_argument(
        "--dpi",
        help="dots per inch to save file",
        type=int,
        default=100,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    plt.rcParams.update({"text.usetex": args.use_latex})
    if args.font_family is not None:
        plt.rcParams.update({"font.family": args.font_family})

    file_list = glob.glob(args.data_regex)

    for file_path in file_list:
        data = pd.read_csv(file_path, delimiter=" ")
        x = data[args.x_value_column_name].to_numpy()
        y = data[args.y_value_column_name].to_numpy()

        plt.scatter(
            x,
            y,
            s=args.scatter_size,
            alpha=args.alpha,
            marker=args.marker,
        )

    plt.xlabel(args.xlabel)
    plt.ylabel(args.ylabel)
    plt.title(args.title)
    plt.tight_layout()
    # Determine output file type and save accordingly
    output_ext = os.path.splitext(args.output.name)[1].lower()
    plt_savefig_kwargs = dict()
    if output_ext == ".png":
        plt_savefig_kwargs["dpi"] = args.dpi
    plt.savefig(args.output.name, bbox_inches="tight", **plt_savefig_kwargs)
    plt.close("all")
